

/*
* Copyright 2023 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the GCG Lab Framework and must not be redistributed.
*/

in VertexData {
#if defined(BLOBBY)
	vec3 object_position;
	vec3 camera_position_object;
#endif
	vec3 position_world;
	vec3 prev_position_world_proj;
	vec3 position_world_proj;
	vec3 normal_world;
	vec2 uv;
} vert;

layout(location=0) out vec4 color;

uniform vec3 camera_world;

uniform vec3 ambientColor;
uniform vec3 diffuseColor;
uniform vec3 specularColor;

uniform float specularAlpha;

uniform mat4 prevViewProjMatrix; // used to calculate the previous position of the vertex (-> movement)
layout(location=1) out vec4 velocity; // the velocity of the vertex

#if !defined(NO_TEX)
uniform sampler2D diffuseTexture;
#endif

#if defined(BLOBBY) || defined(PROC)
	uniform uint time;
#endif

uniform bool draw_normals;
uniform bool draw_texcoords;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;


vec3 phong(vec3 normal, vec3 lightVec, vec3 viewVec, vec3 lightColor, vec3 baseColor, vec3 diffuseColor, vec3 specularColor, float shininess, bool useAttenuation, vec3 attenuationFactors) {
    // Calculate the distance from the light source
    float distance = length(lightVec);

    // Normalize the light vector
    lightVec = normalize(lightVec);

    // Initialize the attenuation coefficient
    float attenuation = 1.0;

    // If attenuation is enabled, compute the attenuation based on distance
    if(useAttenuation) {
        attenuation = 1.0f / (attenuationFactors.x + distance * attenuationFactors.y + distance * distance * attenuationFactors.z);
    }

    // Calculate the reflection vector
    vec3 reflectionVec = reflect(-lightVec, normal);

    // Calculate the diffuse component with channel-specific factors
    float dotNL = max(0, dot(normal, lightVec));
    vec3 diffuse = (baseColor * diffuseColor) * lightColor * dotNL;

    // Calculate the specular component with channel-specific factors
    float dotRV = max(0, dot(reflectionVec, viewVec));
    vec3 specular = specularColor * lightColor * pow(dotRV, shininess);

    // Return the final color, combined with attenuation
    return (diffuse + specular) * attenuation;
}

// Gets the reflected color value from a a certain position, from 
// a certain direction INSIDE of a cornell box of size 3 which is 
// positioned at the origin.
// positionWS:  Position inside the cornell box for which to get the
//              reflected color value for from -directionWS.
// directionWS: Outgoing direction vector (from positionWS towards the
//              outside) for which to get the reflected color value for.
vec3 getCornellBoxReflectionColor(vec3 positionWS, vec3 directionWS) {
	vec3 P0 = positionWS;
	vec3 V  = normalize(directionWS);

	const float boxSize = 1.5;
	vec4[5] planes = vec4[5](
		vec4(-1.0,  0.0,  0.0, -boxSize), // left
		vec4( 1.0,  0.0,  0.0, -boxSize), // right
		vec4( 0.0,  1.0,  0.0, -boxSize), // top
		vec4( 0.0, -1.0,  0.0, -boxSize), // bottom
		vec4( 0.0,  0.0, -1.0, -boxSize)  // back
	);
	vec3[5] colors = vec3[5](
		vec3(1.0, 0.0, 0.0),    // left
		vec3(0.0, 1.0, 0.0),    // right
		vec3(0.96, 0.93, 0.85), // top
		vec3(0.64, 0.64, 0.64), // bottom
		vec3(0.76, 0.74, 0.68)  // back
	);

	for (int i = 0; i < 5; ++i) {
		vec3  N = planes[i].xyz;
		float d = planes[i].w;
		float denom = dot(V, N);
		if (denom <= 0) continue;
		float t = -(dot(P0, N) + d)/denom;
		vec3  P = P0 + t*V;
		float q = boxSize + 0.01;
		if (P.x > -q && P.x < q && P.y > -q && P.y < q && P.z > -q && P.z < q) {
			return colors[i];
		}
	}
	return vec3(0.0, 0.0, 0.0);
}

// Computes the reflection direction for an incident vector I about normal N, 
// and clamps the reflection to a maximum of 180, i.e. the reflection vector
// will always lie within the hemisphere around normal N.
// Aside from clamping, this function produces the same result as GLSL's reflect function.
vec3 clampedReflect(vec3 I, vec3 N)
{
	return I - 2.0 * min(dot(N, I), 0.0) * N;
}

vec3 fresnelSchlick (float cosTheta, vec3 F0) {
    return F0 + (1.0 - F0) * pow (1.0 - cosTheta, 5.0);
}




#if defined(PROC)
	vec4 permute ( vec4 x ) { return mod ((34.0 * x + 1.0) * x , 289.0) ; }

	// Worley noise function by https://github.com/ashima/webgl-noise/blob/master/src/cellular2x2x2.glsl
	// Copyright (C) 2011-2016 by Stefan Gustavson
	// Usage allowed via MIT License
	vec2 cellular2x2x2(vec3 P) {
		#define K 0.142857142857 // 1/7
		#define Ko 0.428571428571 // 1/2-K/2
		#define K2 0.020408163265306 // 1/(7*7)
		#define Kz 0.166666666667 // 1/6
		#define Kzo 0.416666666667 // 1/2-1/6*2
		#define jitter 0.8 // smaller jitter gives less errors in F2
		vec3 Pi = mod(floor(P), 289.0);
		vec3 Pf = fract(P);
		vec4 Pfx = Pf.x + vec4(0.0, -1.0, 0.0, -1.0);
		vec4 Pfy = Pf.y + vec4(0.0, 0.0, -1.0, -1.0);
		vec4 p = permute(Pi.x + vec4(0.0, 1.0, 0.0, 1.0));
		p = permute(p + Pi.y + vec4(0.0, 0.0, 1.0, 1.0));
		vec4 p1 = permute(p + Pi.z); // z+0
		vec4 p2 = permute(p + Pi.z + vec4(1.0)); // z+1
		vec4 ox1 = fract(p1*K) - Ko;
		vec4 oy1 = mod(floor(p1*K), 7.0)*K - Ko;
		vec4 oz1 = floor(p1*K2)*Kz - Kzo; // p1 < 289 guaranteed
		vec4 ox2 = fract(p2*K) - Ko;
		vec4 oy2 = mod(floor(p2*K), 7.0)*K - Ko;
		vec4 oz2 = floor(p2*K2)*Kz - Kzo;
		vec4 dx1 = Pfx + jitter*ox1;
		vec4 dy1 = Pfy + jitter*oy1;
		vec4 dz1 = Pf.z + jitter*oz1;
		vec4 dx2 = Pfx + jitter*ox2;
		vec4 dy2 = Pfy + jitter*oy2;
		vec4 dz2 = Pf.z - 1.0 + jitter*oz2;
		vec4 d1 = dx1 * dx1 + dy1 * dy1 + dz1 * dz1; // z+0
		vec4 d2 = dx2 * dx2 + dy2 * dy2 + dz2 * dz2; // z+1

		d1 = min(d1, d2);
		d1.xy = min(d1.xy, d1.wz);
		d1.x = min(d1.x, d1.y);
		return vec2(sqrt(d1.x));
	}

	// We map the noise to a color gradient between red and yellow
	vec3 proceduralTexture(vec2 uv, vec3 colorA, vec3 colorB, float power) {
		#define PROCEDURAL_SIZE 2.0

		vec2 nsample = cellular2x2x2(vec3(uv * PROCEDURAL_SIZE, time * 0.0005));


		return mix(colorA, colorB, pow(nsample.x, power));
	}
#endif

#if defined(BLOBBY)
	#define MAX_STEPS 100

	// https://github.com/glslify/glsl-smooth-min/blob/master/poly.glsl
	// Copyright (c) 2014 stackgl contributors
	// Usage permitted via MIT License
	float smin(float a, float b, float k) {
		float h = clamp(0.5 + 0.5 * (b - a) / k, 0.0, 1.0);
		return mix(b, a, h) - k * h * (1.0 - h);
	}

	float sphere(vec3 p, vec3 position, float r) {
		return length(p - position) - r;
	}

	#define SCALE 1.7

	float scene(vec3 p) {
		float result =
			sphere(p, vec3(0, SCALE * 0.5 * sin(time*0.001), 0), 0.3 * SCALE);
		result = smin(result,
			sphere(p, vec3(SCALE * 0.4 * cos(time*0.001), SCALE * 0.5 * cos(time*0.001), SCALE * 0.4 * sin(time*0.001)), 0.35 * SCALE),
		0.3);
		result = smin(result,
			sphere(p, vec3(SCALE * 0.35 * sin(time*0.0006), SCALE * 0.3 * cos(time*0.0001), SCALE * 0.35 * sin(time*0.001)), 0.4 * SCALE),
		0.35);
		result = smin(result,
			sphere(p, vec3(SCALE * 0.45 * sin(time*0.008), SCALE * 0.5 * sin(time*0.001), SCALE * 0.3 * sin(time*0.005)), 0.15 * SCALE),
		0.6);
		result = smin(result,
			sphere(p, vec3(SCALE * 0.3 * cos(time*0.004), SCALE * 0.3 * sin(time*0.003), SCALE * 0.35 * cos(time*0.003)), 0.2 * SCALE),
		0.8);

		return result;
	}

	float raymarch(vec3 ro, vec3 rd) {
		float t = 0.0;
		for (int i = 0; i < MAX_STEPS; i++) {
			vec3 p = ro + t * rd;
			float d = scene(p);
			if (d < 0.01) {
				return t;
			}
			t += d;
		}
		return -1.0;
	}

	vec3 getNormal(vec3 p) {
		vec2 e = vec2(0.01, 0.0);
		vec3 n = scene(p) - vec3(
			scene(p - e.xyy),
			scene(p - e.yxy),
			scene(p - e.yyx)
		);

		return normalize(n);
	}
#endif

void main() {
	vec3 n = normalize(vert.normal_world);
	vec3 v = normalize(vert.position_world - camera_world);
	vec3 R = normalize(clampedReflect(v, n));
	vec3 reflectionColor = getCornellBoxReflectionColor(vert.position_world, R);
	vec3 F0 = vec3(0.1); // <-- some kind of plastic
	vec3 reflectivity = fresnelSchlick(dot(n, -v), F0);

	vec3 baseColor = vec3(0.0);
	// Is textured and not procedural
	#if !defined(NO_TEX) && !defined(PROC)
		baseColor += texture(diffuseTexture, vert.uv).rgb * ambientColor;
	#endif

	#if defined(AMB)
		baseColor += ambientColor;
	#endif

	// Is procedural
	#if defined(PROC)
		#if defined(P_LAVA)
			// This lava texture is                     red              and orange
			baseColor = proceduralTexture(vert.uv, vec3(0.7, 0.0, 0.0), vec3(1.0, 1.0, 0.0), 2.0);
		#elif defined(P_WATER)
	        // This water texture is                    blue             and white (caustics)
			baseColor = proceduralTexture(vert.uv, vec3(0.3, 0.53, 1.0), vec3(0.66, 0.95, 1.0), 5.0);
		#else
			// If we don't specify a procedural texture, we use a debug variant
			baseColor = proceduralTexture(vert.uv, vec3(1.0, 0.0, 1.0), vec3(0.0), 1.0);
		#endif
	#endif

	#if defined(BLOBBY)
		vec2 uv_t = vert.uv - 0.5;
		vec3 ro = vert.camera_position_object;
		vec3 object_position = vert.object_position;

		vec3 rd = normalize(object_position - ro);

		float d = raymarch(ro, rd);
		if (d > 0.0) {
			vec3 p = ro + d * rd;
			n = getNormal(p);
			baseColor = vec3(2.0/256, 96.0/256, 98.0/256);
		} else {
			discard;
		}

	#endif


	color = vec4(baseColor, 1); // ambient

	// add directional light contribution
	color.rgb += phong(n, -dirL.direction, -v, dirL.color, baseColor, diffuseColor, specularColor, specularAlpha, false, vec3(0));

	// add point light contribution
	color.rgb += phong(n, pointL.position - vert.position_world, -v, pointL.color, baseColor, diffuseColor, specularColor, specularAlpha, true, pointL.attenuation);


	if (draw_normals) {
		color.rgb = n;
	}
	if (draw_texcoords) {
		color.rgb = vec3(vert.uv, 0);
	}

	velocity = vec4((vert.position_world_proj.xyz * 0.5 + 0.5) - (vert.prev_position_world_proj * 0.5 + 0.5), 1);
}
